@@ -9,8 +9,9 @@ import gzip
 import os
 import shutil
 import struct
-from io import open
+from io import BytesIO, open
 from logging import getLogger
+from typing import IO, Tuple
 
 from Cryptodome.Hash import SHA256
 from cryptography.hazmat.backends import default_backend
@@ -18,9 +19,61 @@ from cryptography.hazmat.primitives import hashes
 
 from .constants import UTF8
 
+logger = getLogger(__name__)
+
 
 class SnowflakeFileUtil(object):
 
+    @staticmethod
+    def get_digest_and_size(src: IO[bytes]) -> Tuple[str, int]:
+        """Gets stream digest and size.
+
+        Args:
+            src: The input stream.
+
+        Returns:
+            Tuple of src's digest and src's size in bytes.
+        """
+        use_openssl_only = os.getenv('SF_USE_OPENSSL_ONLY', 'False') == 'True'
+        CHUNK_SIZE = 16 * 4 * 1024
+        if not use_openssl_only:
+            m = SHA256.new()
+        else:
+            backend = default_backend()
+            chosen_hash = hashes.SHA256()
+            hasher = hashes.Hash(chosen_hash, backend)
+        while True:
+            chunk = src.read(CHUNK_SIZE)
+            if chunk == b'':
+                break
+            if not use_openssl_only:
+                m.update(chunk)
+            else:
+                hasher.update(chunk)
+
+        if not use_openssl_only:
+            digest = base64.standard_b64encode(m.digest()).decode(UTF8)
+        else:
+            digest = base64.standard_b64encode(hasher.finalize()).decode(UTF8)
+
+        size = src.tell()
+        src.seek(0)
+        return digest, size
+
+    @staticmethod
+    def compress_with_gzip_from_stream(src_stream: IO[bytes]) -> Tuple[IO[bytes], int]:
+        """Compresses a stream of bytes with GZIP.
+
+        Args:
+            src_stream: bytes stream
+
+        Returns:
+            A tuple of byte stream and size.
+        """
+        compressed_data = gzip.compress(src_stream.read())
+        src_stream.seek(0)
+        return BytesIO(compressed_data), len(compressed_data)
+
     @staticmethod
     def compress_file_with_gzip(file_name, tmp_dir):
         """Compresses a file with GZIP.
@@ -32,7 +85,6 @@ class SnowflakeFileUtil(object):
         Returns:
             A tuple of gzip file name and size.
         """
-        logger = getLogger(__name__)
         base_name = os.path.basename(file_name)
         gzip_file_name = os.path.join(tmp_dir, base_name + '_c.gz')
         logger.debug('gzip file: %s, original file: %s', gzip_file_name,
@@ -82,7 +134,22 @@ class SnowflakeFileUtil(object):
                     byte = f.read(1)
 
     @staticmethod
-    def get_digest_and_size_for_file(file_name):
+    def get_digest_and_size_for_stream(src_stream: IO[bytes]) -> Tuple[str, int]:
+        """Gets stream digest and size.
+
+        Args:
+            src_stream: The input source stream.
+
+        Returns:
+            Tuple of src_stream's digest and src_stream's size in bytes.
+        """
+        digest, size = SnowflakeFileUtil.get_digest_and_size(src_stream)
+        logger.debug('getting digest and size for stream: %s, %s', digest,
+                     size)
+        return digest, size
+
+    @staticmethod
+    def get_digest_and_size_for_file(file_name: str) -> Tuple[str, int]:
         """Gets file digest and size.
 
         Args:
@@ -91,31 +158,9 @@ class SnowflakeFileUtil(object):
         Returns:
             Tuple of file's digest and file size in bytes.
         """
-        use_openssl_only = os.getenv('SF_USE_OPENSSL_ONLY', 'False') == 'True'
-        CHUNK_SIZE = 16 * 4 * 1024
-        with open(file_name, 'rb') as f:
-            if not use_openssl_only:
-                m = SHA256.new()
-            else:
-                backend = default_backend()
-                chosen_hash = hashes.SHA256()
-                hasher = hashes.Hash(chosen_hash, backend)
-            while True:
-                chunk = f.read(CHUNK_SIZE)
-                if chunk == b'':
-                    break
-                if not use_openssl_only:
-                    m.update(chunk)
-                else:
-                    hasher.update(chunk)
-
-        statinfo = os.stat(file_name)
-        file_size = statinfo.st_size
-        if not use_openssl_only:
-            digest = base64.standard_b64encode(m.digest()).decode(UTF8)
-        else:
-            digest = base64.standard_b64encode(hasher.finalize()).decode(UTF8)
-        logger = getLogger(__name__)
+        digest, size = None, None
+        with open(file_name, 'rb') as src:
+            digest, size = SnowflakeFileUtil.get_digest_and_size(src)
         logger.debug('getting digest and size: %s, %s, file=%s', digest,
-                     file_size, file_name)
-        return digest, file_size
+                     size, file_name)
+        return digest, size
